//
// Bouncer.cpp
//

#include <iostream>
#include <vector>
#include <algorithm>		// for nth_element

// game includes
#include "Ball.h"
#include "Bouncer.h"

// engine includes
#include "Config.h"
#include "DisplayManager.h" 
#include "EventStep.h"
#include "GameManager.h" 
#include "LogManager.h"
#include "WorldManager.h"

// Constructor
Bouncer::Bouncer(bool do_particles, int balls_per_step, int starting_balls,
		 int run_time, int iteration_max) {

  // Set Object attributes.
  setSolidness(df::SPECTRAL);
  registerInterest(df::STEP_EVENT);

  // Set Bouncer attributes.
  m_do_particles = do_particles;
  m_balls_per_step = balls_per_step;
  m_starting_balls = starting_balls;
  m_run_time = run_time;
  m_iteration_max = iteration_max;

  // Setup initial run.
  m_iteration = 0;
  m_bounce_mark = 0;

  // Ready to go.
  m_next_iteration = GM.getStepCount()-1;
  startBounce();
}

// Get bounce ready for next iteration.
void Bouncer::startBounce() {

  // Set up for run by clearing queue.
  for (int i=0; i<LAST; i++)
    m_q[i] = 0;
  m_head = 0;

  // Create number of starting balls.
  createBalls(m_starting_balls);

  // First step.
  m_step_count = 0;
}

// Create indicated number of balls.
void Bouncer::createBalls(int count) {
  if (m_do_particles) {
    for (int i=0; i < count; i++)
      particleBurst();
  } else
    for (int i=0; i < count; i++)
      new Ball;
}

// Handle event.
int Bouncer::eventHandler(const df::Event *p_e) {

  // Every step...
  if (p_e->getType() == df::STEP_EVENT) {

    if (GM.getStepCount() < m_next_iteration)
      return 0;

    if (GM.getStepCount() == m_next_iteration)
      startBounce();

    m_step_count++;

    // Create ball/particles.
    createBalls(m_balls_per_step);

    // Record how long game loop took.
    long int delta = m_clock.delta();
    LM.writeLog(5, "Bouncer::eventHandler(): Delta is %d.", delta);

    // Update queue.
    m_q[m_head] = delta;	   // Add most recent to queue.
    m_head = (m_head + 1) % LAST;  // Circular queue.

    // Compute median.
    std::vector<int> temp_vector;
    for (int i=0; i < LAST; i++)
      temp_vector.push_back(m_q[i]);
    int n = (int) temp_vector.size()/2;
    nth_element (temp_vector.begin(), temp_vector.begin()+n,
		 temp_vector.end());
    int median = temp_vector[n]/1000; // Convert to milliseconds.

    // Write to log.
    df::ObjectList objs = WM.objectsOfType("ball");
    LM.writeLog("BOUNCE: Balls %d. Step %d - %ld of %.0f msec ( median %d )",
		objs.getCount(),
		m_step_count, delta/1000, 
		TARGET_LIMIT, median);

    // If done, stop iteration.
    if (m_run_time) {
      if (m_step_count > m_run_time) {
	stopBounce();
      }
    } else {
      if (median >= TARGET_LIMIT)
	stopBounce();
    }
    return 1;
  }

  // If we get here, we have ignored this event.
  return 0;
}

// Stop this iteration. If this is third iteration, stop the game.
void Bouncer::stopBounce() {
  static int tally = 0;		// count total

  // Update stats for this iteration.
  m_iteration++;

  // Ball count went THRESHOLD past the limit, so subtract.
  // Also, median of LAST balls was used, so subtract 1/2 that.

  // Balls at median.
  int balls = m_starting_balls
              + (m_step_count * m_balls_per_step)
              - (LAST/2) * m_balls_per_step;

  // Subtract THRESHOLD.
  balls = balls * (1 - (int) THRESHOLD);
  tally += balls;
  LM.writeLog("BOUNCE: Iteration %d - balls: %d", m_iteration, balls);

  // If this is final iteration, shut things down.
  if (m_iteration == m_iteration_max) {
    m_bounce_mark = tally/m_iteration_max;
    if (m_do_particles) {
      std::string s = df::Config::getInstance().match("burst");
      if (s != "")
	m_bounce_mark *= atoi(s.c_str());
    }
    LM.writeLog("BOUNCE: Average max %s: %d", 
		m_do_particles ? "particles" : "objects", m_bounce_mark);
    GM.setGameOver();
    return;
  }

  // Remove all balls.
  df::ObjectList objs = WM.getAllObjects();
  df::ObjectListIterator i(&objs);
  for (i.first(); !i.isDone(); i.next()) {
    df::Object *p_o = i.currentObject();
    if (p_o->getType() == "ball" || p_o->getType() == "Particle") 
      WM.markForDelete(p_o);
  }

  // Ready to go again.
  // (Wait for one step to allow WM.markForDelete() to take effect.)
  m_next_iteration = GM.getStepCount() + 1;
}

// Return average number of objects supported (bounce-mark).
int Bouncer::getBounceMark() {
  return m_bounce_mark;
}

// Add particle burst.
void Bouncer::particleBurst() {
  std::string s;

  // Defaults if no match in df-config.txt.
  int number = 50;
  int number_spread = 0;
  float size = 1.0f;
  float size_spread = 0.0f;
  float speed = 0.01f;
  float speed_spread = 0.0f;
  float age = 100000.0f;  // Make large so particles don't expire.
  int age_spread = 0;
  unsigned char opacity = 255;
  int opacity_spread = 0;
  df::Vector direction(0,0);
  float direction_spread = 0.0f;
  float position_spread = 0.0f;
  int r = 255;
  int g = 255;
  int b = 255;
  unsigned char color_spread = 0;
  
  // Number of particles.
  s = df::Config::getInstance().match("burst");
  if (s != "")
    number = atoi(s.c_str());
  s = df::Config::getInstance().match("burst-spread");
  if (s != "")
    number_spread = atoi(s.c_str());
  
  // Speed.
  s = df::Config::getInstance().match("speed");
  if (s != "")
    speed = (float) atof(s.c_str());
  s = df::Config::getInstance().match("speed-spread");
  if (s != "")
    speed_spread = (float) atof(s.c_str());
  
  // Size.
  s = df::Config::getInstance().match("size");
  if (s != "")
    size = (float) atof(s.c_str());
  s = df::Config::getInstance().match("size-spread");
  if (s != "")
    size_spread = (float) atof(s.c_str());
  
  // Age.
  s = df::Config::getInstance().match("age");
  if (s != "")
    age = (float) atoi(s.c_str());
  s = df::Config::getInstance().match("age-spread");
  if (s != "")
    age_spread = atoi(s.c_str());
  
  // Opacity.
  s = df::Config::getInstance().match("opacity");
  if (s != "")
    opacity = atoi(s.c_str());
  s = df::Config::getInstance().match("opacity-spread");
  if (s != "")
    opacity_spread = atoi(s.c_str());
  
  // Color.
  s = df::Config::getInstance().match("color-r");
  if (s != "")
    r = atoi(s.c_str());
  s = df::Config::getInstance().match("color-g");
  if (s != "")
    g = atoi(s.c_str());
  s = df::Config::getInstance().match("color-b");
  if (s != "")
    b = atoi(s.c_str());
  s = df::Config::getInstance().match("color-spread");
  if (s != "")
    color_spread = atoi(s.c_str());
  
  // Direction.
  s = df::Config::getInstance().match("direction-x");
  if (s != "")
    direction.setX((float) atof(s.c_str()));
  s = df::Config::getInstance().match("direction-y");
  if (s != "")
    direction.setY((float) atof(s.c_str()));
  s = df::Config::getInstance().match("direction-spread");
  if (s != "")
    direction_spread = (float) atof(s.c_str());
  
  // Position spread.
  s = df::Config::getInstance().match("position-spread");
  if (s != "")
    position_spread = (float) atof(s.c_str());
  
  // Location.
  df::Vector cell(DM.getHorizontal()/2.0f, DM.getVertical()/2.0f);

  // Add particles.
  df::addParticles(number, number_spread, 
		   cell, position_spread,
		   direction, direction_spread, 
		   size, size_spread,
		   speed, speed_spread,
		   (int) age, (int) age_spread, 
		   opacity, opacity_spread,
		   r, g, b, color_spread);

  return;
}
